/*
 * Genital
 * Bart Trzynadlowski
 * Public domain
 */


/*
 * genital.c: Main program, memory space, RAM, and ROM code
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "68k.h"
#include "genital.h"
#include "gfx.h"
#include "keyboard.h"
#include "misc.h"
#include "save.h"
#include "vdp.h"
#include "vga.h"
#include "z80.h"

#define LINE_CLOCK_68K  488
#define LINE_CLOCK_Z80  (LINE_CLOCK_68K / 2)


struct genesis_struct   genesis;
struct control_struct   control;
struct profile_struct   profile;

int                     fskip = 0, fskip_count = 0;
int                     save_state = 0, load_state = 0, debug = 0;


void ShowHelp()
{
        printf("Genital Version %s by Bart Trzynadlowski: A Free Sega Genesis"
               " Emulator\n", VERSION);
        printf("Usage:     genital <SMD or BIN ROM> [options]\n");
        printf("Options:   -?,-h      Show this help text\n");
        printf("           -info      Show information about the ROM\n");
        printf("           -dump      Dump RAM, VRAM, CRAM, and VSRAM on exit\n");
        printf("           -star      Use the Starscream 68000 emulator\n");
        printf("           -raze      Use the RAZE Z80 emulator\n");
        printf("           -fskip #   Frames to skip before drawing and blitting\n");
        printf("           -vsync     Uses vsync to impose a speed limit\n");
        printf("           -mode13    320x200 Mode 0x13 resolution\n");
        printf("           -modex     320x240 ModeX resolution\n");
        printf("           -usa       American Sega Genesis system\n");
        printf("           -europe    European Sega MegaDrive system\n");
        printf("           -japan     Japanese Sega MegaDrive system\n");
        printf("           -z80       Attempt to actually emulate the Z80\n");
        printf("           -hint      Allow horizontal interrupts\n");
        printf("           -nopri     No Scroll/Sprite priorities\n");
        printf("           -sprlimit  Enable sprite limiting\n");
        exit(0);
}

int FindOption(char *option, int argc, char **argv)
{
        int     i;
        
        /* argv[1] = rom; options start at argv[2] */
        for (i = 1; i < argc; i++)
        {                
                if (strcmp(argv[i], option) == 0)
                        return i;       /* found it! return position */
        }

        return 0;
}


/*
 * Misc. code
 */

void SwapBuffer(unsigned char *buffer, int length)
{
        int     i, j;

        /* swap bytes in each word */
        for (i = 0; i < length; i += 2)
        {
                j = buffer[i];
                buffer[i] = buffer[i + 1];
                buffer[i + 1] = j;
        }
}


/*
 * ROM code
 */

void ConvertSMDToBIN(unsigned char *bin_block, unsigned char *smd_block)
{
        int     i, o = 1, e = 0;

        /* convert 16KB of SMD to BIN */
        for (i = 0; i < 8192; i++)
        {
                bin_block[o] = smd_block[i];                
                bin_block[e] = smd_block[i+8192];
                o += 2;
                e += 2;
        }
}        

void ROMLoad(char *file)
{
        FILE            *fp;
        int             i, j = 0, num_blocks, size;
        unsigned char   smd_block[16384];

        if ((fp = fopen(file, "rb")) == NULL)
        {
                fprintf(stderr, "Genital: File could not be opened: %s\n",
                        file);
                exit(1);
        }
        if ((genesis.rom = (unsigned char *) calloc(4194304 + 4,
        sizeof(unsigned char))) == NULL)
        {
                fprintf(stderr, "Genital: Failed to allocate 4MB RAM for " \
                        "ROM\n");
                exit(1);
        }
        
        /* get size of file, and number of blocks in case it is SMD */
        fseek(fp, 0, SEEK_END);
        fgetpos(fp, (void *) &size);
        num_blocks = (size - 512) / 16384;
        fseek(fp, 0, SEEK_SET);
       
        /* load the file */
        if (strstr(file, ".bin") != NULL)       /* BIN */
        {
                if (size > 4194304)     /* too big? */
                {
                        fprintf(stderr, "Genital: ROM file excedes 4MB\n");
                        exit(1);
                }
                else
                        fread(genesis.rom, 1, size, fp);
        }
        else if (strstr(file, ".smd") != NULL)  /* SMD */
        {
                fseek(fp, 512, SEEK_SET);       /* skip SMD header */
                for (i = 0; i < num_blocks; i++)
                {
                        fread(smd_block, 1, 16384, fp);
                        ConvertSMDToBIN(&genesis.rom[j], smd_block);
                        j += 16384;     /* next 16KB chunk in rom[] */
                }
        }
        else
        {
                fprintf(stderr, "Genital: Unrecognized ROM format\n");
                exit(1);
        }

        fclose(fp);
}        

void ROMUnload()
{
        free(genesis.rom);
}

void ShowROMInfo(char *file)
{
        FILE            *fp;
        unsigned char   header[16384];
        unsigned char   smd[16384];
        char            string[49];

        /* check to see if it a BIN or SMD file */
        if ((strstr(file, ".bin") != NULL) || (strstr(file, ".smd") != NULL))
        {
                /* open file and read the header */
                if ((fp = fopen(file, "rb")) == NULL)
                {
                        fprintf(stderr, "Genital: File could not be opened: " \
                                "%s\n", file);
                        exit(1);
                }

                if (strstr(file, ".bin") != NULL)       /* bin file */
                        fread(header, 1, 512, fp);
                else                                    /* smd file */
                {
                        /* get first block of SMD */
                        fread(smd, 1, 512, fp);         /* skip header */
                        fread(smd, 1, 16384, fp);
                        ConvertSMDToBIN(header, smd);   /* decode */
                }

                printf("Showing information from cartridge header...\n");
                printf("                    File: %s\n", file);
                string[16] = '\0';
                strncpy(string, &header[0x100], 16);
                printf("                  System: %s\n", string);
                strncpy(string, &header[0x110], 16);
                printf("               Copyright: %s\n", string);
                string[48] = '\0';
                strncpy(string, &header[0x120], 48);
                printf("    Game name (domestic): %s\n", string);
                strncpy(string, &header[0x150], 48);
                printf("    Game name (overseas): %s\n", string);
                if (header[0x180] == 'G' && header[0x181] == 'M')
                        printf("           Software type: game\n");
                else if (header[0x180] == 'A' && header[0x181] == 'l')
                        printf("           Software type: educational\n");
                else
                        printf("           Software type: %c%c\n",
                               header[0x180], header[0x181]);
                string[11] = '\0';
                strncpy(string, &header[0x183], 11);
                printf("Product code and version: %s\n", string);
                printf("                Checksum: %02X%02X\n", header[0x18e],
                       header[0x18f]);
                string[16] = '\0';
                strncpy(string, &header[0x190], 16);
                printf("             I/O support: %s\n", string);
                printf("       ROM start address: %02X%02X%02X%02X\n",
                       header[0x1a0], header[0x1a1], header[0x1a2],
                       header[0x1a3]);
                printf("         ROM end address: %02X%02X%02X%02X\n",
                       header[0x1a4], header[0x1a5], header[0x1a6],
                       header[0x1a7]);
                string[20] = '\0';
                strncpy(string, &header[0x1bc], 20);                
                printf("              Modem data: %s\n", string);
                string[40] = '\0';
                strncpy(string, &header[0x1c8], 40);
                printf("                    Memo: %s\n", string);
                string[3] = '\0';
                strncpy(string, &header[0x1f0], 3);
                printf("               Countries: %s\n", string);

                fclose(fp);
        }
        else
        {
                fprintf(stderr, "Genital: Unrecognized ROM format\n");
                exit(1);
        }
        exit(0);
}


/*
 * RAM code and memory (RAM, ROM) access handlers
 */

unsigned MemReadByte(unsigned addr)
{
        if (addr >= 0xe00000)  
                return genesis.ram[(addr ^ 1) & 0xffff];
        if (addr < 0x400000)
                return genesis.rom[addr ^ 1];
        return 0;
}

unsigned MemReadWord(unsigned addr)
{
        if (addr >= 0xe00000)
                return GetWord(genesis.ram + (addr & 0xfffe));
        else if (addr < 0x400000)
                return GetWord(genesis.rom + addr);
        return 0;
}

void MemWriteByte(unsigned addr, unsigned data)
{
        if (addr >= 0xe00000)   /* RAM */
                genesis.ram[(addr ^ 1) & 0xffff] = data;
}

void RAMInit()
{
        if ((genesis.ram = calloc(65536, sizeof(unsigned char))) == NULL)
        {
                fprintf(stderr, "Genital: Failed to allocate 64KB of RAM " \
                        "for work RAM\n");
                exit(1);
        }
}

void RAMDeinit()
{
        free(genesis.ram);
}


/*
 * System I/O Registers (including Z80 control ports and RAM)
 */

unsigned IOReadByte(unsigned addr)
{
        if (addr == 0xa10002 || addr == 0xa10003)   /* control pad 1 */
        {
                if (control.ctrl1_select == 1)
                        return (control.ctrl1 & 0xff);
                else
                        return (control.ctrl1 >> 8);
        }
        if (addr == 0xa10005)   /* control pad 2 */
                return 0xff;
        if (addr == 0xa10001)   /* version */
                return genesis.version;

        if (addr == 0xa11100)   /* Z80 BUSREQ */
        {
                if (genesis.busreq)     /* 0 means busreq (Z80 inactive) */
                        return 0;
                else
                        return 1;
        }

        /*
         * Z80 area
         */

        if ((addr >= 0xa00000) && (addr < 0xa08000))
        {
                if (genesis.busreq)     /* important: _only_ during busreq */
                        return Z80Read(addr & 0x7fff);
        }

        return 0;
}

unsigned IOReadWord(unsigned addr)
{
        if (addr == 0xa10002)   /* control pad 1 (word) */
        {
                control.ctrl1 &= 0xffff;
                if (control.ctrl1_select == 1)
                        return ((control.ctrl1 & 0xff) << 8) | (control.ctrl1 & 0xff);
                else
                        return (control.ctrl1 & 0xff00) | ((control.ctrl1 >> 8) & 0xff);
        }

        if (addr == 0xa11100)   /* Z80 BUSREQ */
        {
                if (genesis.busreq)     /* 0 means busreq (Z80 inactive) */
                        return 0;
                else
                        return 0x100;
        }

        return 0;
}

void IOWriteByte(unsigned addr, unsigned data)
{        
        if (addr == 0xa10002 || addr == 0xa10003)   /* control pad 1 */
        {
                if (data & 0x40)        /* select high? */
                        control.ctrl1_select = 1;
                else
                        control.ctrl1_select = 0;
                return;
        }

        /*
         * Z80 BUSREQ -- words are written to 0xa11100, I'm assuming bytes
         * would be written since we are modifying word bit 8.
         */

        if (addr == 0xa11100)
        {
                genesis.busreq = data & 1;      /* bit 8 (bit 0 of byte) */
                return;
        }

        /*
         * Z80 reset
         */

        if (addr == 0xa11200)
        {
                if (!(data & 1))        /* writing zero resets */
                        Z80Reset();
                return;
        }

        /*
         * Z80 RAM
         */

        if ((addr >= 0xa00000) && (addr < 0xa08000))
        {
                if (genesis.busreq)     /* important: _only_ during busreq */
                        Z80Write(addr & 0x7fff, data);
        }
}

void IOWriteWord(unsigned addr, unsigned data)
{
        if (addr == 0xa10002)   /* control pad 1 (word) */
        {
                if (data & 0x40 || data & 0x4000)       /* select high? */
                        control.ctrl1_select = 1;
                else
                        control.ctrl1_select = 0;
                return;
        }

        /*
         * Z80 BUSREQ
         */

        if (addr == 0xa11100)
        {               
                if (data & 0x100)
                        genesis.busreq = 1;
                else
                        genesis.busreq = 0;
                return;
        }

        /*
         * Z80 reset
         */

        if (addr == 0xa11200)
        {
                if (!(data & 0x100))    /* writing zero to bit 8 resets */
                        Z80Reset();
                return;
        }
}

/*
 * Genesis emulator
 */

int main(int argc, char **argv)
{        
        FILE    *fp;
        int     i;

        if (KeyboardInit())
        {
                KeyboardDeinit();
                printf("Genital: Failed to install keyboard driver\n");
                exit(1);
        }

        /* command line */
        if (argc == 1)                          /* show help */
                ShowHelp();
        if (FindOption("-?", argc, argv) || FindOption("-h", argc, argv))
                ShowHelp();
        if (FindOption("-info", argc, argv))    /* show ROM info */
                ShowROMInfo(argv[1]);
        if ((fskip = FindOption("-fskip", argc, argv)))
        {                
                fskip = fskip_count = atoi(argv[fskip + 1]);
                if (fskip < 0)  /* if below 0, make it at least 0 */
                        fskip = fskip_count = 0;
        }       
        if (FindOption("-usa", argc, argv))     /* which country */
                genesis.version = 0xa0;
        else if (FindOption("-europe", argc, argv))
                genesis.version = 0xe0;
        else if (FindOption("-japan", argc, argv))
                genesis.version = 0x20;
        else
                genesis.version = 0x80;         /* default is USA */
        if (FindOption("-vsync", argc, argv))   /* wait until vsync? */
                genesis.vsync = 1;
        else
                genesis.vsync = 0;
        if (FindOption("-nopri", argc, argv))   /* priorities? */
                genesis.pri = 0;
        else
                genesis.pri = 1;
        if (FindOption("-sprlimit", argc, argv))        /* sprite limiting? */
                genesis.spr_limit = 1;
        else
                genesis.spr_limit = 0;
        if (FindOption("-hint", argc, argv))    /* h-ints? */
                genesis.hint = 1;
        else
                genesis.hint = 0;
        if (FindOption("-z80", argc, argv))     /* z80? */
                genesis.z80_emu = 1;
        else
                genesis.z80_emu = 0;

        genesis.busreq = 1;
        genesis.save_slot = 0;

        /* start... */
        ROMLoad(argv[1]);        
        RAMInit();
        /* Z80 is initialized before the 68K to provide a proper z80_ram */
        if (FindOption("-raze", argc, argv))
                Z80Init(RAZE_Z80);
        else
                Z80Init(DEFAULT_Z80);
        if (FindOption("-star", argc, argv))
                CPUInit(STARSCREAM_CPU);
        else
                CPUInit(DEFAULT_CPU);
        VDPInit();

        if (FindOption("-mode13", argc, argv))
                GFXInit(MODE13);
        else if (FindOption("-modex", argc, argv))
                GFXInit(MODEX);
        else
                GFXInit(MODE13);
        while (!control.quit)
        {
                VDPArmHIntTimer();
                for (i = 0; i < 224; i++)
                {                                                
                        VDPUpdateHVCounter(i);
                        VDPHInterrupt();
                        Z80Run(LINE_CLOCK_Z80);
                        CPURun(LINE_CLOCK_68K);
                        if (fskip_count == 0)
                                DrawLine(i);
                }
                VDPVBlankStart();
                Z80Interrupt(0xff);
                for (i = i; i < 224 + 38; i++)
                {
                        VDPUpdateHVCounter(i);
                        Z80Run(LINE_CLOCK_Z80);
                        CPURun(LINE_CLOCK_68K);
                }
                if (fskip_count == 0)
                        BlitScreen();
                VDPVBlankEnd();

                /* update fskip counter */
                fskip_count--;
                if (fskip_count < 0)
                        fskip_count = fskip;    /* reload */

                if (save_state)
                {
                        GSTSave(argv[1]);
                        save_state = 0;         /* reset flag */
                }
                if (load_state)
                {
                        GSTLoad(argv[1]);       
                        load_state = 0;         /* reset flag */
                }
                if (debug)
                {
                        CPUDebug();
                        debug = 0;
                }
        }
        GFXDeinit();
        KeyboardDeinit();

        /* dump memory */
        if (FindOption("-dump", argc, argv))
        {
                if ((fp = fopen("ram.out", "wb")) != NULL)
                {
                        SwapBuffer(genesis.ram, 65536);
                        fwrite(genesis.ram, 1, 65536, fp);
                        fclose(fp);
                        printf("Genital: RAM dumped to RAM.OUT\n");
                }
                else
                        printf("Genital: Failed to dump RAM\n");
                if ((fp = fopen("vram.out", "wb")) != NULL)
                {
                        SwapBuffer(vdp.vram, 65536);
                        fwrite(vdp.vram, 1, 65536, fp);
                        fclose(fp);
                        printf("Genital: VRAM dumped to VRAM.OUT\n");
                }
                else
                        printf("Genital: Failed to dump VRAM\n");
                if ((fp = fopen("cram.out", "wb")) != NULL)
                {
                        SwapBuffer(vdp.cram, 128);
                        fwrite(vdp.cram, 1, 128, fp);
                        fclose(fp);
                        printf("Genital: CRAM dumped to CRAM.OUT\n");
                }
                else
                        printf("Genital: Failed to dump CRAM\n");
                if ((fp = fopen("vsram.out", "wb")) != NULL)
                {
                        SwapBuffer(vdp.vsram, 80);
                        fwrite(vdp.vsram, 1, 80, fp);
                        fclose(fp);
                        printf("Genital: VSRAM dumped to VSRAM.OUT\n");
                }
                else
                        printf("Genital: Failed to dump VSRAM\n");
        }

        /* quit */
        RAMDeinit();
        ROMUnload();
        VDPDeinit();
       
        return 0;
}


