/*
 * gfx.c: Graphics rendering engine
 */

#include <pc.h>
#include <stdio.h>
#include <stdlib.h>
#include <sys/farptr.h>
#include "genital.h"
#include "gfx.h"
#include "misc.h"
#include "tile.h"
#include "vdp.h"
#include "vga.h"

/*
 * Screen is 320x224. The video buffer has each line as: 8 + 320 + 8. There
 * are also 8 lines above and below the virtual screen. These unused 8 pixel
 * tall or wide areas are for h- and v-scrolling.
 */

#define BUFFER_SIZE     ((8 + 320 + 8) * 224) + ((8 + 320 + 8) * (8 + 8))
#define BUFFER_CELL_WIDTH       ((8 + 320 + 8) >> 3)
unsigned char           *video_buffer;  /* ~80640 bytes */

void             (*BlitScreen)(int);


void DrawSprites(int line, int pr)      /* Genital code */
{
        int             i, tile, x, y, yline, xlimit, height, tile_x, tile_y;
        unsigned char   *video, *spr;

        for (i = vdp.sprite_count - 1; i >= 0; --i)
        {                
                spr = &vdp.sprite_addr[vdp.sprite_order[i] << 3];
                tile = GetWord(spr + 4);

                /* don't draw if incorrect priority */
                if (((tile & 0x8000) >> 15) != pr)
                        goto loop_end;

                x = (GetWord(spr + 6) & 0x1ff) - 0x80;
                y = (GetWord(spr + 0) & 0x1ff) - 0x80;

                /* if (line - y) >= 0, then sprite is above or on this line
                   and may be visible. a negative sprite means the sprite is
                   not on the screen */
                if ((yline = line - y) < 0)
                        goto loop_end;

                xlimit = ((spr[2^1] << 1) & 0x18) + x;
                height = (spr[2^1] & 0x3);

                /* do if on this line */
                if ((xlimit > -8) && (x < 320) && (yline >= 0) && (yline <= (height << 3) + 7))
                {
                        tile_y = yline & 7;     /* line = tile line */

                        /* modify the actual tile entry to get the proper
                           vertical tile */                        
                        if (tile & 0x1000)      /* vflip */
                                tile += height - (yline >> 3);
                        else
                                tile += (yline >> 3);
                        height++;                                               

                        if (tile & 0x800)       /* hflip */
                        {                                                        
                                video = (line * ((BUFFER_CELL_WIDTH) << 3)) + video_buffer + xlimit;
                                for (tile_x = xlimit; tile_x >= x; tile_x -= 8)
                                {
                                        if ((tile_x > -8) && (tile_x < 320))                                        
                                                DrawSpriteTile(tile, tile_y, video);
                                        tile += height;
                                        video -= 8; 
                                }
                        }
                        else
                        {
                                video = (line * ((BUFFER_CELL_WIDTH) << 3)) + video_buffer + x;
                                for (tile_x = x; tile_x <= xlimit; tile_x += 8)
                                {
                                        if ((tile_x > -8) && (tile_x < 320))
                                                DrawSpriteTile(tile, tile_y, video);
                                        tile += height;
                                        video += 8;
                                }
                        }
                }
        loop_end:
        }
        
}

/*
 * Note: Window code is an ugly hack. I need to implement info found on p60 of
 * sega2.doc to get this working correctly in Raiden Trad and other games.
 * Calculation of vdp.window_x, vdp.window_y, etc. is done by the register
 * handling code in vdp.c
 */

void DrawWindow(int line)
{        
        int             i;
        unsigned char   *ntab, *video;

        /* only do if on line */
        if (!(((line >> 3) >= vdp.window_y) && ((line >> 3) < (vdp.window_y + vdp.window_height))))
                return;

        video = (line * ((BUFFER_CELL_WIDTH) << 3)) + video_buffer + (vdp.window_x << 3);
        ntab = vdp.window_addr + (((line >> 3) * vdp.window_width) << 1);

        for (i = 0; i < vdp.window_width; i++)
        {
                DrawTile((line & 7), ntab, video);
                ntab += 2;
                video += 8;
        }               
}

void DrawLine(int line)
{
        if (!(vdp.reg[1] & 0x40))       /* display is off, don't draw */
                return;

        VDPCalcSpriteOrder(line);

        if (genesis.pri)
        {
                #define PRI 1

                /* Scroll B low */
                #define SCROLL 1
                #define SOLID 1
                        #include "gfx_rend.h"
                #undef SOLID
                #undef SCROLL

                /* Scroll A low */
                #define SCROLL 0
                #define SOLID 1
                        #include "gfx_rend.h"
                #undef SOLID
                #undef SCROLL

                DrawSprites(line, 0);

                /* Scroll B high */
                #define SCROLL 1
                #define SOLID 0
                        #include "gfx_rend.h"
                #undef SOLID
                #undef SCROLL

                /* Scroll A high */
                #define SCROLL 0
                #define SOLID 0
                        #include "gfx_rend.h"
                #undef SOLID
                #undef SCROLL

                DrawWindow(line);
                DrawSprites(line, 1);

                #undef PRI
        }
        else
        {
                #define PRI 0

                /* Scroll B */
                #define SCROLL 1
                #define SOLID 1
                        #include "gfx_rend.h"
                #undef SOLID
                #undef SCROLL

                /* Scroll A low */
                #define SCROLL 0
                #define SOLID 0
                        #include "gfx_rend.h"
                #undef SOLID
                #undef SCROLL

                DrawWindow(line);
                DrawSprites(line, 0);
                DrawSprites(line, 1);

                #undef PRI
        }
} 

void Mode13BlitScreen()
{
        int             i;
        unsigned int    pc_vram = 0xa0000;
        unsigned char   *video;

        video = video_buffer;

        if (genesis.vsync)      /* wait until vsync */
                while (!(inportb(INPUT_STATUS) & 8))
                        ;
                
        if (vdp.reg[12] & 0x1)  /* H40 */
        {
                for (i = 0; i < 200; i++)
                {
                        FastMemcpy((void *) video, (void *) pc_vram, 320);
                        pc_vram += 320;         /* to next line in PC VRAM */
                        video += 320 + 8 + 8;   /* ... in video buffer */
                }
        }
        else                    /* H32 */
        {
                for (i = 0; i < 200; i++)
                {
                        FastMemset((void *) pc_vram, 255, 32);
                        pc_vram += 32;          /* center screen */
                        FastMemcpy((void *) video, (void *) pc_vram, 256);
                        pc_vram += 256;
                        FastMemset((void *) pc_vram, 255, 32);
                        pc_vram += 32;          /* next line... */
                        video += 320 + 8 + 8;
                }

        }
}

void ModeXBlitScreen()
{
        int             i, j, k;
        unsigned        pc_vram = 0xa0000;
        unsigned char   *video;

        video = video_buffer;

        if (genesis.vsync)      /* wait until vsync */
                while (!(inportb(INPUT_STATUS) & 8))
                        ;
                
        /*
         * ModeX:
         *
         * This blitting routine is slow and ugly, but this is my first time
         * working with ModeX and I can't think of a better way.
         */

        if (vdp.reg[12] & 0x1)  /* H40 */
        {
                for (i = 0; i < 224; i++)
                {
                        for (j = 0; j < 4; j++)
                        {
                                /* plane */
                                outportb(0x3c4, 0x02);
                                outportb(0x3c5, 0x01 << j);
                                /* start at 0, every 4 bytes because we're
                                   doing one plane at a time */
                                for (k = j; k < 320; k += 4)
                                {
                                        _farnspokeb(pc_vram, video[k]);
                                        pc_vram++;
                                }
                                pc_vram -= 320 / 4; /* reset for next plane */
                        }
                        pc_vram += 320 / 4;
                        video += 320 + 8 + 8;   /* next line */
                }
        }
        else                    /* H32 */
        {
                for (i = 0; i < 224; i++)
                {
                        for (j = 0; j < 4; j++)
                        {
                                /* plane */
                                outportb(0x3c4, 0x02);
                                outportb(0x3c5, 0x01 << j);
                                /* start at 0, every 4 bytes because we're
                                   doing one plane at a time */
                                for (k = 0; k < 32; k += 4)
                                {
                                        _farnspokeb(pc_vram, 255);
                                        pc_vram++;
                                }
                                for (k = j; k < 256; k += 4)
                                {
                                        _farnspokeb(pc_vram, video[k]);
                                        pc_vram++;
                                }
                                for (k = 0; k < 32; k += 4)
                                {
                                        _farnspokeb(pc_vram, 255);
                                        pc_vram++;
                                }
                                pc_vram -= 320 / 4; /* reset for next plane */
                        }
                        pc_vram += 320 / 4;
                        video += 320 + 8 + 8;   /* next line */
                }
        }
}

void GFXSwitch(int res)
{
        if (res == MODE13)
                BlitScreen = &Mode13BlitScreen;
        else
                BlitScreen = &ModeXBlitScreen;

        GraphicsInit(res);
}


void GFXInit(int res)
{
        if ((video_buffer = calloc(BUFFER_SIZE, sizeof(unsigned char)))
            == NULL)
        {
                fprintf(stderr, "Genital: GFXInit() failed\n");
                exit(1);
        }

        /* forward 8 lines and 8 pixels to the start of the displayed
           portion of the buffer */
        video_buffer += ((8 + 320 + 8) * 8);// + 8;

        GFXSwitch(res);
}

void GFXDeinit()
{
        video_buffer -= ((8 + 320 + 8) * 8);// + 8;
        free(video_buffer);

        GraphicsDeinit();
}




