;
; misc.asm: CPU identification, profiling, memory block manipulation, etc.
;

%define no_underscores 0        ; if you need a _ prefix for global or extern
                                ; items, keep this as 0 (COFF, OBJ); if you
                                ; do not need one, set to 1 (ELF)
%define sensitive_regs 0        ; if your C compiler needs routines to
                                ; preserve registers on their own, define as
                                ; 1, otherwise keep as 0
                              

%if no_underscores = 0
        %define ASSERTCPU       _AssertCPU
        %define STARTCOUNTING   _StartCounting
        %define ENDCOUNTING     _EndCounting
        %define FASTMEMSET      _FastMemset
        %define FASTMEMCPY      _FastMemcpy
        %define FPUMEMCPY       _FPUMemcpy        
%elif no_underscores = 1
        %define ASSERTCPU       AssertCPU
        %define STARTCOUNTING   StartCounting
        %define ENDCOUNTING     EndCounting
        %define FASTMEMSET      FastMemset
        %define FASTMEMCPY      FastMemcpy
        %define FPUMEMCPY       FPUMemcpy
%endif

global ASSERTCPU
global STARTCOUNTING
global ENDCOUNTING
global FASTMEMSET
global FASTMEMCPY
global FPUMEMCPY

%define cache_align times ($$-$) & 31 nop       ; align to 32-byte boundary

bits 32
section .data

        cache_align
        start_tsc       dd 0, 0                 ; save TSC value here

bits 32
section .text

; register saving macros: if sensitive_regs = 0, does not save/restore
%macro save_regs 0
%if sensitive_regs = 1
        pushad
%elif sensitive_regs = 0
%endif
%endmacro
%macro restore_regs 0
%if sensitive_regs = 1
        popad
%elif sensitive_regs = 0
%endif
%endmacro


; AssertCPU(): Returns non-zero if system has a processor that is TSC capable
cache_align
ASSERTCPU:
        push    ebp
        mov     ebp, esp

        save_regs

        ; if there is no CPUID, this processor is assumed not to be usable
        pushf
        pop     ebx             ; flags
        mov     eax, ebx        ; stick flags here
        or      eax, 0x200000   ; set ID bit
        push    eax
        popf                    ; now our modified flags are here
        pushf
        pop     eax             ; get them back to eax
        test    eax, 0x200000   ; is the ID bit set?
        jnz     .cpuid_cpu      ; yes! we have a processor with CPUID!

.bad_cpu:
        xor     eax, eax        ; return 0

.good_cpu:
        restore_regs

        pop     ebp
ret

.cpuid_cpu:
        mov     eax, 1          ; get feature flags from CPUID
        cpuid
        test    edx, 0x10       ; time stamp counter supported?
        jz      .bad_cpu

        mov     eax, 1

jmp     .good_cpu                

; StartCounting(): Start the timestamp counter
cache_align
STARTCOUNTING:
        push    ebp
        mov     ebp, esp

        save_regs

        ; get the timestamp counter
        rdtsc
        mov     [start_tsc], eax
        mov     [start_tsc+4], edx

        restore_regs

        pop     ebp
ret

; EndCounting(element): Stop counting, adds the low dword to where the
; "element" pointer is pointing
cache_align
ENDCOUNTING:
        push    ebp
        mov     ebp, esp

        save_regs

        ; get the timestamp counter and find the difference
        mov     ebx, [ebp+8]            ; get element pointer
        rdtsc
        sub     eax, [start_tsc]
        sbb     edx, [start_tsc+4]
        add     [ebx], eax              ; lower 32-bits of difference

        restore_regs

        pop     ebp
ret

; FastMemset(mem, value, bytes): Memory fill
cache_align
FASTMEMSET:
        push    ebp
        mov     ebp, esp

        save_regs

        push    es
        mov     ax, fs
        mov     es, ax          ; destination uses global selector

        mov     edi, [ebp+8]    ; region to fill
        mov     al, [ebp+12]    ; value to fill w/
        mov     ecx, [ebp+16]   ; number of bytes
        rep     stosb
                     
        pop     es

        restore_regs

        pop     ebp
ret

; FastMemcpy(src, dest, bytes): Dword-based memory copy
cache_align
FASTMEMCPY:
        push    ebp
        mov     ebp, esp

        save_regs

        push    es
        mov     ax, fs
        mov     es, ax          ; destination uses global selector

        mov     esi, [ebp+8]    ; get source address
        mov     ecx, [ebp+16]   ; number of bytes
        mov     edi, [ebp+12]   ; get destination address
        shr     ecx, 2          ; divide by 4 for dwords
        rep     movsd           ; copy

        pop     es

        restore_regs

        pop     ebp
ret

; FPUMemcpy(src, dest, bytes): Qword-based memory copy (assumes FPU present)
cache_align
FPUMEMCPY:
        push    ebp
        mov     ebp, esp

        save_regs

        push    es
        mov     ax, fs
        mov     es, ax          ; destination uses global selector

        mov     esi, [ebp+8]    ; get source address
        mov     ecx, [ebp+16]   ; number of bytes
        mov     edi, [ebp+12]   ; get destination address
        shr     ecx, 3          ; divide by 8 for qwords

.copy:
        fild    qword [esi]
        fistp   qword [es:edi]
        add     esi, 8
        add     edi, 8
        dec     ecx
        jnz     .copy

        pop     es

        restore_regs

        pop     ebp
ret

