/*
 * save.c: Save state management - thanks to Charles Mac Donald for his doc
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "68k.h"
#include "genital.h"
#include "vdp.h"
#include "vga.h"
#include "z80.h"

/*
 * GST Notes:
 *
 * - GST files are 140408 bytes in size
 * - Registers are stored in Intel format
 * - All RAM areas are stored in Motorola format (except CRAM apparently)
 */

/*
 * Special thanks to Charles Mac Donald for his excellent GST document, and
 * to Dave for DGen.
 */

void GSTSave(char *file)
{
        FILE            *fp;
        char            *ext[] = { ".gs0", ".gs1", ".gs2", ".gs3", ".gs4",
                                   ".gs5", ".gs6", ".gs7", ".gs8", ".gs9" };
        unsigned char   *gst;
        int             i;

        if ((fp = fopen(strcat(strtok(file, "."), ext[genesis.save_slot]), "wb")) == NULL)
                return;

        if ((gst = (unsigned char *) calloc(140408, sizeof(unsigned char))) == NULL)
                return;

        gst[0] = 'G';
        gst[1] = 'S';
        gst[2] = 'T';
        gst[6] = 0xe0;
        gst[7] = 0x40;

        /* save 68K regs -- remember to /4 for dword pointers */
        for (i = 0; i < 8; i++)
                *((unsigned *) gst + ((0x80 + i*4) / 4)) = CPUGetD(i);
        /* save 68K Ax regs */
        for (i = 0; i < 8; i++)
                *((unsigned *) gst + ((0xa0 + i*4) / 4)) = CPUGetA(i);
        /* save PC and SR */        
        *((unsigned *) gst + 0xc8 / 4) = CPUGetPC();
        *((unsigned *) gst + 0xd0 / 4) = CPUGetSR();
        /* save 68K RAM */
        for (i = 0; i < 65536; i++)
                gst[0x2478 + i] = genesis.ram[i];
        SwapBuffer(&gst[0x2478], 65536);        /* convert to big endian */
        
        /* save VDP registers */
        for (i = 0; i < 24; i++)
                gst[0xfa + i] = vdp.reg[i];
        /* save CRAM -- not byteswapped (fairly certain) */
        for (i = 0; i < 128; i++)
                gst[0x112 + i] = vdp.cram[i];
        /* save VSRAM */
        for (i = 0; i < 80; i++)
                gst[0x192 + i] = vdp.vsram[i]; 
        SwapBuffer(&gst[0x192], 80);            /* convert to big endian */
        /* save VRAM */
        for (i = 0; i < 65536; i++)
                gst[0x12478 + i] = vdp.vram[i];
        SwapBuffer(&gst[0x12478], 65536);       /* convert to big endian */

        /* save Z80 registers */
        *((unsigned *) gst + 0x404 / 4) = Z80GetAF();
        *((unsigned *) gst + 0x408 / 4) = Z80GetBC();
        *((unsigned *) gst + 0x40c / 4) = Z80GetDE();
        *((unsigned *) gst + 0x410 / 4) = Z80GetHL();
        *((unsigned *) gst + 0x414 / 4) = Z80GetIX();
        *((unsigned *) gst + 0x418 / 4) = Z80GetIY();
        *((unsigned *) gst + 0x41c / 4) = Z80GetPC();
        *((unsigned *) gst + 0x420 / 4) = Z80GetSP();
        *((unsigned *) gst + 0x424 / 4) = Z80GetAFPrime();
        *((unsigned *) gst + 0x428 / 4) = Z80GetBCPrime();
        *((unsigned *) gst + 0x42c / 4) = Z80GetDEPrime();
        *((unsigned *) gst + 0x430 / 4) = Z80GetHLPrime();
        *((unsigned *) gst + 0x434 / 4) = Z80GetI();
        /* save Z80 RAM */
        for (i = 0; i < 8192; i++)
                gst[0x474 + i] = genesis.z80_ram[i];

        fwrite(gst, 1, 140408, fp);
        fclose(fp);
        free(gst);
}

void GSTLoad(char *file)
{
        FILE            *fp;
        char            *ext[] = { ".gs0", ".gs1", ".gs2", ".gs3", ".gs4",
                                   ".gs5", ".gs6", ".gs7", ".gs8", ".gs9" };
        unsigned char   *gst;
        int             i;

        if ((fp = fopen(strcat(strtok(file, "."), ext[genesis.save_slot]), "rb")) == NULL)
                return;

        if ((gst = (unsigned char *) calloc(140408, sizeof(unsigned char))) == NULL)
                return;
        fread(gst, 1, 140408, fp);      /* read it into gst buffer */
        fclose(fp);

        /* get 68K Dx regs  -- remember to /4 for dword pointers */
        for (i = 0; i < 8; i++)
                CPUSetD(i, *((unsigned *) gst + ((0x80 + i*4) / 4)));
        /* get 68K Ax regs */
        for (i = 0; i < 8; i++)
                CPUSetA(i, *((unsigned *) gst + ((0xa0 + i*4) / 4)));
        /* get PC and SR */        
        CPUSetPC(*((unsigned *) gst + 0xc8 / 4));
        CPUSetSR(*((unsigned *) gst + 0xd0 / 4));
        /* get 68K RAM */
        for (i = 0; i < 65536; i++)
                genesis.ram[i] = gst[0x2478 + i];
        SwapBuffer(genesis.ram, 65536);
        
        /* get VDP registers */
        for (i = 0; i < 24; i++)
                VDPSetReg(i, gst[0xfa + i]);
        /* get CRAM -- not byteswapped (fairly certain) */
        for (i = 0; i < 128; i++)
                vdp.cram[i] = gst[0x112 + i];
        ResetPalette();
        /* get VSRAM */
        for (i = 0; i < 80; i++)
                vdp.vsram[i] = gst[0x192 + i];
        SwapBuffer(vdp.vsram, 80);
        /* get VRAM */
        for (i = 0; i < 65536; i++)
                vdp.vram[i] = gst[0x12478 + i];
        SwapBuffer(vdp.vram, 65536);

        /* get Z80 registers */
        Z80SetAF(*((unsigned *) gst + 0x404 / 4));
        Z80SetBC(*((unsigned *) gst + 0x408 / 4));
        Z80SetDE(*((unsigned *) gst + 0x40c / 4));
        Z80SetHL(*((unsigned *) gst + 0x410 / 4));
        Z80SetIX(*((unsigned *) gst + 0x414 / 4));
        Z80SetIY(*((unsigned *) gst + 0x418 / 4));
        Z80SetPC(*((unsigned *) gst + 0x41c / 4));
        Z80SetSP(*((unsigned *) gst + 0x420 / 4));
        Z80SetAFPrime(*((unsigned *) gst + 0x424 / 4));
        Z80SetBCPrime(*((unsigned *) gst + 0x428 / 4));
        Z80SetDEPrime(*((unsigned *) gst + 0x42c / 4));
        Z80SetHLPrime(*((unsigned *) gst + 0x430 / 4));
        Z80SetI(*((unsigned *) gst + 0x434 / 4));
        /* get Z80 RAM */
        for (i = 0; i < 8192; i++)
                genesis.z80_ram[i] = gst[0x474 + i];

        free(gst);
}


        
