;
; tile.asm: Tile drawing routines
;     

%define no_underscores 0        ; if you need a _ prefix for global or extern
                                ; items, keep this as 0 (COFF, OBJ); if you
                                ; do not need one, set to 1 (ELF)
%define sensitive_regs 0        ; if your C compiler needs routines to
                                ; preserve registers on their own, define as
                                ; 1, otherwise keep as 0
                              

%if no_underscores = 0
        %define DRAWTILE        _DrawTile
        %define DRAWTILESOLID   _DrawTileSolid
        %define DRAWSPRITETILE  _DrawSpriteTile
        %define VIDEO_BUFFER    _video_buffer
        %define VDP             _vdp
%elif no_underscores = 1
        %define DRAWTILE        DrawTile
        %define DRAWTILESOLID   DrawTileSolid
        %define DRAWSPRITETILE  DrawSpriteTile
        %define VIDEO_BUFFER    video_buffer
        %define VDP             vdp
%endif

global DRAWTILE
global DRAWTILESOLID
global DRAWSPRITETILE
extern VIDEO_BUFFER
extern VDP

%define cache_align times ($$-$) & 31 nop       ; align to 32-byte boundary

; vdp access: keep in sync with vdp.h. all members can now be accessed as:
; [VDP+vdpstruc.member_name], special %defines for registers are provided
struc vdpstruc
        .vram                   resd 1
        .cram                   resd 1
        .vsram                  resd 1
        .ctrlport_mode          resd 1
        .access_mode            resd 1
        .addr                   resd 1
        .dma_mode               resd 1
        .control_port           resd 1
        .status_register        resd 1
        .hv_counter             resd 1
        .reg                    resd 24
        .scroll_a_addr          resd 1
        .scroll_b_addr          resd 1
        .window_addr            resd 1
        .sprite_addr            resd 1
        .hscroll_addr           resd 1
        .window_x               resd 1
        .window_y               resd 1
        .window_width           resd 1
        .window_height          resd 1
        .sprite_order           resd 80
        .sprite_count           resd 1
        .hint_timer             resd 1
endstruc

; register saving macros: if sensitive_regs = 0, does not save/restore
%macro save_regs 0
%if sensitive_regs = 1
        pushad
%elif sensitive_regs = 0
%endif
%endmacro
%macro restore_regs 0
%if sensitive_regs = 1
        popad
%elif sensitive_regs = 0
%endif
%endmacro

; plotting macros: eax = pattern, dh = palette, dl = scratch, edi =
; vdeo_buffer 
%macro plot_nohflip 0 
        rol     eax, 4
        mov     dl, al
        and     dl, 0xf
        jz      %%skip
        or      dl, dh
        mov     [edi], dl
%%skip:
        inc     edi
%endmacro
%macro plot_hflip 0
        mov     dl, al
        and     dl, 0xf
        jz      %%skip
        or      dl, dh
        mov     [edi], dl
%%skip:
        shr     eax, 4
        inc     edi
%endmacro
%macro plot_nohflip_solid 0     ; uses bg color in bl rather than transparency
        rol     eax, 4
        mov     dl, al
        and     dl, 0xf
        jz      %%zero
        or      dl, dh
        mov     [edi], dl
        inc     edi
        jmp     %%end
%%zero:
        mov     [edi], bl       ; plot the bg color
        inc     edi
%%end:
%endmacro
%macro plot_hflip_solid 0
        mov     dl, al
        and     dl, 0xf
        jz      %%zero
        or      dl, dh
        mov     [edi], dl
        shr     eax, 4
        inc     edi
        jmp     %%end
%%zero:
        mov     [edi], bl       ; plot the bg color
        shr     eax, 4
        inc     edi
%%end:
%endmacro

; void DrawTile(int line, unsigned char *tile, unsigned char *destination):
; Draws one line of a Scroll layer tile.
cache_align
DRAWTILE:
        push    ebp
        mov     ebp, esp

        save_regs

        mov     ebx, [ebp+12]   ; name table address
        mov     edi, [ebp+16]   ; destination video buffer
        mov     ebp, [ebp+8]    ; get line #

        mov     ax, [ebx]       ; get entry
        mov     dx, ax
        and     dh, 0x60
        shr     dh, 1           ; dh = 0000 00PP, P = palette
        test    eax, 0x1000     ; vflip?
        jz      .no_vflip

        mov     esi, ebp        ; displacement for the scanline (w/ vflip)
        and     esi, 7
        xor     esi, 7
        shl     esi, 2
        jmp     .continue

cache_align
.no_vflip:
        mov     esi, ebp        ; displacement for the scanline
        and     esi, 7
        shl     esi, 2

.continue:
        test    eax, 0x800      ; hflip?
        jnz     near .hflip     ; yes? go to the hflip blitter

        and     eax, 0x7ff      ; index * 32 + vram = pattern line
        add     esi, [VDP+vdpstruc.vram]
        shl     eax, 5
        add     esi, eax        ; add to displacement

        mov     ax, [esi]       ; get 8 pixels
        shl     eax, 16
        add     esi, 2
        mov     ax, [esi]
        cmp     eax, 0
        jz      .no_plot_0      ; don't waste time if clear
        plot_nohflip            ; plot them w/out hflip
        plot_nohflip
        plot_nohflip
        plot_nohflip
        plot_nohflip
        plot_nohflip
        plot_nohflip
        plot_nohflip

.no_plot_0:
        restore_regs

        pop     ebp
ret

.hflip:        
        and     eax, 0x7ff      ; index * 32 + vram = pattern line
        add     esi, [VDP+vdpstruc.vram]
        shl     eax, 5
        add     esi, eax        ; add to displacement

        mov     ax, [esi]       ; get 8 pixels
        shl     eax, 16
        add     esi, 2
        mov     ax, [esi]
        cmp     eax, 0
        jz      .no_plot_1      ; don't waste time if clear
        plot_hflip              ; plot them w/out hflip
        plot_hflip
        plot_hflip
        plot_hflip
        plot_hflip
        plot_hflip
        plot_hflip
        plot_hflip

.no_plot_1:
        restore_regs

        pop     ebp
ret

; void DrawTileSolid(int line, unsigned char *tile,
;                    unsigned char *destination):
; Draws one line of a Scroll layer tile, does not check for transparency.
cache_align
DRAWTILESOLID:
        push    ebp
        mov     ebp, esp

        save_regs

        mov     ebx, [ebp+12]   ; name table address
        mov     edi, [ebp+16]   ; destination video buffer
        mov     ebp, [ebp+8]    ; get line #

        mov     ax, [ebx]       ; get entry
        mov     dx, ax
        and     dh, 0x60
        shr     dh, 1           ; dh = 0000 00PP, P = palette
        test    eax, 0x1000     ; vflip?
        jz      .no_vflip

        mov     esi, ebp        ; displacement for the scanline (w/ vflip)
        and     esi, 7
        xor     esi, 7
        shl     esi, 2
        jmp     .continue

cache_align
.no_vflip:
        mov     esi, ebp        ; displacement for the scanline
        and     esi, 7
        shl     esi, 2

.continue:
        ; use ebx to get bg color for 0
        mov     ebx, [VDP+vdpstruc.reg+(7*4)]
        and     bl, 0x3f        ; keep within 6 bits

        test    eax, 0x800      ; hflip?
        jnz     near .hflip     ; yes? go to the hflip blitter

        and     eax, 0x7ff      ; index * 32 + vram = pattern line
        add     esi, [VDP+vdpstruc.vram]
        shl     eax, 5
        add     esi, eax        ; add to displacement

        mov     ax, [esi]       ; get 8 pixels
        shl     eax, 16
        add     esi, 2
        mov     ax, [esi]
        plot_nohflip_solid      ; plot them w/out hflip
        plot_nohflip_solid 
        plot_nohflip_solid
        plot_nohflip_solid
        plot_nohflip_solid
        plot_nohflip_solid
        plot_nohflip_solid
        plot_nohflip_solid
        
        restore_regs

        pop     ebp
ret

.hflip:        
        and     eax, 0x7ff      ; index * 32 + vram = pattern line
        add     esi, [VDP+vdpstruc.vram]
        shl     eax, 5
        add     esi, eax        ; add to displacement

        mov     ax, [esi]       ; get 8 pixels
        shl     eax, 16
        add     esi, 2
        mov     ax, [esi]
        plot_hflip_solid        ; plot them w/ hflip
        plot_hflip_solid
        plot_hflip_solid
        plot_hflip_solid
        plot_hflip_solid
        plot_hflip_solid
        plot_hflip_solid
        plot_hflip_solid

        restore_regs

        pop     ebp
ret

; void DrawSpriteTile(unsigned tile, int line, unsigned char *destination):
; Draws one line of a sprite tile, does not check for transparency.
cache_align
DRAWSPRITETILE:
        push    ebp
        mov     ebp, esp

        save_regs

        mov     eax, [ebp+8]    ; tile descriptor
        mov     edi, [ebp+16]   ; destination video buffer
        mov     ebp, [ebp+12]   ; line #

        mov     dx, ax
        and     dh, 0x60
        shr     dh, 1           ; dh = 0000 00PP, P = palette
        test    eax, 0x1000     ; vflip?
        jz      .no_vflip

        mov     esi, ebp        ; displacement for the scanline (w/ vflip)
        and     esi, 7
        xor     esi, 7
        shl     esi, 2
        jmp     .continue

cache_align
.no_vflip:
        mov     esi, ebp        ; displacement for the scanline
        and     esi, 7
        shl     esi, 2

.continue:
        test    eax, 0x800      ; hflip?
        jnz     near .hflip     ; yes? go to the hflip blitter

        and     eax, 0x7ff      ; index * 32 + vram = pattern line
        add     esi, [VDP+vdpstruc.vram]
        shl     eax, 5
        add     esi, eax        ; add to displacement

        mov     ax, [esi]       ; get 8 pixels
        shl     eax, 16
        add     esi, 2
        mov     ax, [esi]
        plot_nohflip            ; plot them w/out hflip
        plot_nohflip
        plot_nohflip
        plot_nohflip
        plot_nohflip
        plot_nohflip
        plot_nohflip
        plot_nohflip
        
        restore_regs

        pop     ebp
ret

.hflip:        
        and     eax, 0x7ff      ; index * 32 + vram = pattern line
        add     esi, [VDP+vdpstruc.vram]
        shl     eax, 5
        add     esi, eax        ; add to displacement

        mov     ax, [esi]       ; get 8 pixels
        shl     eax, 16
        add     esi, 2
        mov     ax, [esi]
        plot_hflip              ; plot them w/ hflip
        plot_hflip
        plot_hflip
        plot_hflip
        plot_hflip
        plot_hflip
        plot_hflip
        plot_hflip

        restore_regs

        pop     ebp
ret



