/*
 * z80.c: Z80 emulation abstraction interface (processor and memory)
 */

#include <stdio.h>
#include <stdlib.h>
#include "genital.h"
#include "raze/raze.h"
#include "z80.h"

int             z80_type = DEFAULT_Z80;

int             (*Z80Run)(int);
int             (*Z80Interrupt)(int);

int             bank = 0;       /* bank reg data */

/*
 * Z80 read/write
 */

unsigned Z80Read(unsigned addr)
{
        unsigned        window_addr;

        if (addr < 0x2000)      /* Z80 RAM */
                return genesis.z80_ram[addr];
        if (addr >= 0x8000)     /* 68K window */
        {        
                window_addr = (bank << 15) + (addr & 0x7fff);
                if ((window_addr < 0x400000) || (window_addr >= 0xe00000))
                        return MemReadByte(window_addr);
                else    /* try the IO area, no use in VDP or anything else */
                        return IOReadByte(window_addr);
        }

        return 0;
}

void Z80Write(unsigned addr, unsigned data)
{
        unsigned        window_addr;

        if (addr < 0x2000)      /* Z80 RAM */
                genesis.z80_ram[addr] = data;
        if (addr == 0x6000)     /* bank register */
        {
                bank >>= 1;     /* make room for top bit */
                bank |= (data & 1) << 8;
                bank &= 0x1ff;
                return;
        }

        if (addr >= 0x8000)     /* 68K window */
        {
                window_addr = (bank << 15) + (addr & 0x7fff);
                if (window_addr >= 0xe00000)
                        MemWriteByte(window_addr, data);
                else    /* try the IO area, no use in VDP or anything else */
                        IOWriteByte(window_addr, data);
                return;
        }
}


/*
 * RAZE functions
 */

unsigned char RAZERead(unsigned short addr)
{       
        return Z80Read(addr); 
}

void RAZEWrite(unsigned short addr, unsigned char data)
{
        Z80Write(addr, data);
}                
                
unsigned char RAZEFakeRead(unsigned short addr)
{
        return 0;
}

void RAZEFakeWrite(unsigned short addr, unsigned char data)
{
}

int RAZERun(int cycles)
{
        if (genesis.busreq)
                return 0;
        if (genesis.z80_emu)
                z80_emulate(cycles);
        return 0;
}

int RAZEInterrupt(int bus)
{
        if (genesis.busreq)
                return 0;
        if (genesis.z80_emu)
        {
                z80_raise_IRQ(bus);
                z80_emulate(32);
                z80_lower_IRQ();
        }
        return 0;
}

unsigned Z80GetAF()
{
        if (z80_type == RAZE_Z80)
                return z80_get_reg(Z80_REG_AF);
        return 0;
}

void Z80SetAF(unsigned data)
{
        if (z80_type == RAZE_Z80)
                z80_set_reg(Z80_REG_AF, data);
}

unsigned Z80GetBC()
{
        if (z80_type == RAZE_Z80)
                return z80_get_reg(Z80_REG_BC);
        return 0;
}

void Z80SetBC(unsigned data)
{
        if (z80_type == RAZE_Z80)
                z80_set_reg(Z80_REG_BC, data);
}

unsigned Z80GetDE()
{
        if (z80_type == RAZE_Z80)
                return z80_get_reg(Z80_REG_DE);
        return 0;
}

void Z80SetDE(unsigned data)
{
        if (z80_type == RAZE_Z80)
                z80_set_reg(Z80_REG_DE, data);
}

unsigned Z80GetHL()
{
        if (z80_type == RAZE_Z80)
                return z80_get_reg(Z80_REG_HL);
        return 0;
}

void Z80SetHL(unsigned data)
{
        if (z80_type == RAZE_Z80)
                z80_set_reg(Z80_REG_HL, data);
}

unsigned Z80GetIX()
{
        if (z80_type == RAZE_Z80)
                return z80_get_reg(Z80_REG_IX);
        return 0;
}

void Z80SetIX(unsigned data)
{
        if (z80_type == RAZE_Z80)
                z80_set_reg(Z80_REG_IX, data);
}

unsigned Z80GetIY()
{
        if (z80_type == RAZE_Z80)
                return z80_get_reg(Z80_REG_IY);
        return 0;
}

void Z80SetIY(unsigned data)
{
        if (z80_type == RAZE_Z80)
                z80_set_reg(Z80_REG_IY, data);
}

unsigned Z80GetPC()
{
        if (z80_type == RAZE_Z80)
                return z80_get_reg(Z80_REG_BC);
        return 0;
}

void Z80SetPC(unsigned data)
{
        if (z80_type == RAZE_Z80)
                z80_set_reg(Z80_REG_PC, data);
}

unsigned Z80GetSP()
{
        if (z80_type == RAZE_Z80)
                return z80_get_reg(Z80_REG_SP);
        return 0;
}

void Z80SetSP(unsigned data)
{
        if (z80_type == RAZE_Z80)
                z80_set_reg(Z80_REG_SP, data);
}

unsigned Z80GetAFPrime()
{
        if (z80_type == RAZE_Z80)
                return z80_get_reg(Z80_REG_AF2);
        return 0;
}

void Z80SetAFPrime(unsigned data)
{
        if (z80_type == RAZE_Z80)
                z80_set_reg(Z80_REG_AF2, data);
}

unsigned Z80GetBCPrime()
{
        if (z80_type == RAZE_Z80)
                return z80_get_reg(Z80_REG_BC2);
        return 0;
}

void Z80SetBCPrime(unsigned data)
{
        if (z80_type == RAZE_Z80)
                z80_set_reg(Z80_REG_BC2, data);
}


unsigned Z80GetDEPrime()
{
        if (z80_type == RAZE_Z80)
                return z80_get_reg(Z80_REG_DE2);
        return 0;
}

void Z80SetDEPrime(unsigned data)
{
        if (z80_type == RAZE_Z80)
                z80_set_reg(Z80_REG_DE2, data);
}

unsigned Z80GetHLPrime()
{
        if (z80_type == RAZE_Z80)
                return z80_get_reg(Z80_REG_HL2);
        return 0;
}

void Z80SetHLPrime(unsigned data)
{
        if (z80_type == RAZE_Z80)
                z80_set_reg(Z80_REG_HL2, data);
}

unsigned Z80GetI()
{
        if (z80_type == RAZE_Z80)
                return z80_get_reg(Z80_REG_IR);
        return 0;
}

void Z80SetI(unsigned data)
{
        if (z80_type == RAZE_Z80)
                z80_set_reg(Z80_REG_IR, data);
}

void Z80Init(int type)
{
        z80_type = type;

        if (type == RAZE_Z80)           /* RAZE core */
        {
                /* allocate 8K Z80 RAM */
                if ((genesis.z80_ram = (unsigned char *) calloc(8192, sizeof(unsigned char))) == NULL)
                {
                        printf("Genital: Failed to allocate 8K of RAM for Z80 RAM\n");
                        exit(1);
                }

                /* Set up memory map */
                z80_init_memmap();

                z80_map_fetch(0x0000, 0x1fff, genesis.z80_ram);
                z80_map_read(0x0000, 0x1fff, genesis.z80_ram);
                z80_map_write(0x0000, 0x1fff, genesis.z80_ram);
                z80_add_write(0x2000, 0xffff, Z80_MAP_HANDLED, &RAZEWrite);
                z80_add_read(0x2000, 0xffff, Z80_MAP_HANDLED, &RAZERead);
                z80_set_in(&RAZEFakeRead);
                z80_set_out(&RAZEFakeWrite);

                z80_end_memmap();
                z80_reset();

                Z80Run = &RAZERun;
                Z80Interrupt = &RAZEInterrupt;
        }
        else
        {
                printf("Genital: Unknown Z80 emulator type: %d\n", type);
                exit(1);
        }
}

void Z80Reset()
{
        if (z80_type == RAZE_Z80)
                z80_reset();
}        




